
"""

   m can be
   
       n-1/2 for clarinets
       n for most instruments 
   
   for n integer

"""

import math

SPEED_OF_SOUND = 346.1   # 25C
                         # 30C is 349.0

semitone = {
   'C' : 0,
   'D' : 2,
   'E' : 4,
   'F' : 5,
   'G' : 7,
   'A' : 9,
   'B' : 11
}

def fqc(note):
    if not isinstance(note, str): return note
    
    s = semitone[note[0].upper()]
    note = note[1:]
    if note.startswith('b'):
        s -= 1
        note = note[1:]
    if note.startswith('#'):
        s += 1
        note = note[1:]
    
    s += 12 * int(note)
    
    return 440.0 * 2**((s-57)/12.0)



def cot(x):
    return math.cos(x) / math.sin(x)

def area_of_hole_of_diameter(x):
    r = x / 2.0
    return math.pi * (r*r)

def length_of_fqc(fqc, m):
    return SPEED_OF_SOUND / fqc * m * 0.5

def fqc_of_length(length, m):
    return SPEED_OF_SOUND / length * m * 0.5

def end_flange_length_correction(outer_diameter, inner_diameter):
    a = inner_diameter / 2.0
    w = (outer_diameter - inner_diameter) / 2.0
    
    return a * (0.821 - 0.13*((0.42+w/a)**-0.54))

def hole_length_correction(hole_diameter, bore_diameter):
    # As per p.63-64
    outer_correction = 0.7
    inner_correction = 1.3 - 0.9 * hole_diameter/bore_diameter
    a = hole_diameter / 2.0

    result = a * (inner_correction+outer_correction)
    return result

# def closed_hole_length_correction(hole_diameter, bore_diameter, wall_thickness):
#     """ From flute spreadsheet. 
#         Tiny. """
#    return 0.25 * wall_thickness * (hole_diameter/bore_diameter)**2


def func(left_length, right_length, hole_length, substitution_length,
         bore_area, hole_area,
         m):
    """ Solve for zero by varying substitution_length """
    
    k = m * math.pi / substitution_length
    
    return (
        bore_area * cot(k*left_length - k*substitution_length) +
        bore_area * cot(k*right_length) +
        hole_area * cot(k*hole_length)
    )

def length_with_hole(left_length, right_length, hole_length,
                     bore_area, hole_area, m):
    f = lambda x: func(left_length, right_length, hole_length, x,
                       bore_area, hole_area, m)

    # *Cautiously* look for a solution near the end of the pipe        
    step = 1e-3   #Initially step by 1mm

    #Initial guess
    x = left_length+right_length
    assert f(x) > 0.0, 'Something is weird, are your inputs sane?'
    
    while step > 1e-10:        
        while f(x) > 0.0:
            x -= step
        x += step    
        step /= 2.0

    return x


def position_hole(determinants, hole_length, bore_area):
    """ Returns distance from left hand end of tube """

    x = max([ item[1] for item in determinants ])
    for i in xrange(100):
        errors = [ ]
        for m, desired_length, length, hole_area in determinants:
            errors.append( length_with_hole(x, length-x, hole_length, bore_area, hole_area, m) \
                           - desired_length )
        error = sum(errors)/len(errors)
        
        if abs(error) < 1e-10: break
        
        x -= error
    
    assert error < 1e-5, 'Hole positioning failed to converge'
    #print ', '.join(['%.1fmm / %.1f%%' % (-errors[i]*1000.0,-errors[i]/determinants[i][1]*100.0) for i in xrange(len(determinants))])
    
    return x


def improve_positions(scorer, positions):
    step = 0.001
    
    score = scorer(positions)
    while step >= 0.00005:
        #print 'Step:', step, math.sqrt(score)
        any = False
        for i in xrange(len(positions)):
            new_positions = positions[:]
            new_positions[i] = positions[i] + step
            new_score = scorer(new_positions)
            if new_score < score: 
                positions = new_positions
                score = new_score
                any = True
                continue

            new_positions[i] = positions[i] - step
            new_score = scorer(new_positions)
            if new_score < score: 
                positions = new_positions
                score = new_score
                any = True
    
        if not any:
            step *= 0.5

    return positions


def design_instrument(
        base_note,
        base_m,
        outer_diameter_mm,
        inner_diameter_mm,
        hole_diameters_mm,
        fingerings):
    """
        Place the finger holes of a cylindrical instrument.
        
        hole spec is a list of tuples defining the holes from the end inward. 
        The tuples consist of:
            the desired note
            the m of the desired note
            the diameter of the hole in mm
            a list indicating which of the previous holes are closed when determining the hole position
        
        What is the m value?
        
            Instruments with a mouthpiece that behaves like a closed end of a pipe (reeds, lips),
            should use an m value of 0.5 (1.5, 2.5, 3.5... for higher registers).
        
            Instruments with a mouthpiece that behaves like an open end of a pipe (flutes, recorders),
            should use an m value of 1 (2, 3, 4... for higher registers).
        
    """
    
    base_fqc = fqc(base_note)
    
    outer_diameter = outer_diameter_mm / 1000.0
    inner_diameter = inner_diameter_mm / 1000.0
    
    wall_thickness = (outer_diameter-inner_diameter)*0.5
    
    base_length = length_of_fqc(base_fqc, base_m)
    end_correction = end_flange_length_correction(outer_diameter, inner_diameter)
    
    print 'Wall thickness    %8.1fmm' % (wall_thickness * 1000.0)
    print 'End correction    %8.1fmm' % (end_correction * 1000.0)
    print
    print 'Instrument length %8.1fmm' % ((base_length-end_correction)*1000.0)
    print
    print 'Note: instrument length includes an end correction but not a mouthpiece length correction'
    print ' - make a pipe about this length, then trim it based on measured frequency'
    print ' - hole positions below are measured from the *end* of the pipe'
    
    print
    print 'Positioning holes'

    hole_diameters = [ item / 1000.0 for item in hole_diameters_mm ]
    hole_positions = [ ]
        
    #for desired_note, m, hole_diameter_mm, holes_closed in hole_spec:
    for i in xrange(len(hole_diameters)):
        determinants = [ ]
        
        for m, desired_note, fingering in fingerings:
            if fingering[i]: continue
            any_open = False
            for j in xrange(i+1, len(hole_diameters)):
                if not fingering[j]:
                    any_open = True
            if any_open: continue
            
            desired_fqc = fqc(desired_note)
            desired_length = length_of_fqc(desired_fqc, m)

            current_length = base_length
            for j in xrange(i):
                if not fingering[j]:
                    current_length = length_with_hole(
                        hole_positions[j],
                        current_length - hole_positions[j],
                        wall_thickness +
                            hole_length_correction(hole_diameters[j], inner_diameter),
                        area_of_hole_of_diameter(inner_diameter),
                        area_of_hole_of_diameter(hole_diameters[j]),
                        m)
            
            determinants.append( 
               (m, desired_length, current_length,
                area_of_hole_of_diameter(hole_diameters[i])) ) 
        
        assert determinants, 'No note determines hole %d' % (i+1)
            
        #desired_fqc = fqc(desired_note)
        #desired_length = length_of_fqc(desired_fqc, m)
        #
        #assert len(holes_closed) == len(holes_so_far)
        #
        #current_length = base_length
        #for i in xrange(len(holes_so_far)):
        #    #if holes_closed[i]:
        #    #    correction = closed_hole_length_correction(
        #    #        holes_so_far[i][1], 
        #    #        inner_diameter,
        #    #        wall_thickness)
        #    #    #print 'closed hole correction', correction #Tiny
        #    #    current_length += correction
        #    if not holes_closed[i]:
        #        current_length = length_with_hole(
        #            holes_so_far[i][0],
        #            current_length - holes_so_far[i][0],
        #            wall_thickness +
        #                hole_length_correction(holes_so_far[i][1], inner_diameter),
        #            area_of_hole_of_diameter(inner_diameter),
        #            area_of_hole_of_diameter(holes_so_far[i][1]),
        #            m)
        
        hole_position = position_hole(
            determinants,
            wall_thickness + 
                hole_length_correction(hole_diameters[i], inner_diameter),  
            area_of_hole_of_diameter(inner_diameter))
        
        hole_positions.append( hole_position )



    def evaluate(hole_positions):
        errors = [ ]
        for m, desired_note, fingering in fingerings:
            desired_fqc = fqc(desired_note)
            desired_length = length_of_fqc(desired_fqc, m)

            current_length = base_length
            for j in xrange(len(hole_diameters)):
                if not fingering[j]:
                    current_length = length_with_hole(
                        hole_positions[j],
                        current_length - hole_positions[j],
                        wall_thickness +
                            hole_length_correction(hole_diameters[j], inner_diameter),
                        area_of_hole_of_diameter(inner_diameter),
                        area_of_hole_of_diameter(hole_diameters[j]),
                        m)
            
            errors.append( (current_length - desired_length)/desired_length )
        
        return errors

    def score(hole_positions):
        return sum( x*x for x in evaluate(hole_positions) ) 

    print 'Improving hole positions'
    hole_positions = improve_positions(score, hole_positions)

    print
    print 'Errors:'
    print
    
    errors = evaluate(hole_positions)
    for i in xrange(len(fingerings)):
        print '%4s %6.1f%%'%(fingerings[i][1], errors[i]*100.0)

    print
    print 'RMS Error: %.2f%%' % (math.sqrt(sum( x*x for x in errors )/len(errors)) * 100.0)    
    print

    print
    print 'Hole    Hole     Distance   Inter-hole'
    print '      diameter   from end     spacing'
    print

    prev_pos = None
    for i in xrange(len(hole_diameters)):
        dist_from_end = base_length-end_correction - hole_positions[i]
    
        if prev_pos is not None:
            print ' '*28, '%6.1fmm' % ((dist_from_end-prev_pos)*1000.0)
        prev_pos = dist_from_end
    
        print '%4d %6.1fmm %8.1fmm' % (i+1, hole_diameters[i]*1000.0, dist_from_end*1000.0)

    print



